/*
 * TrainRI.java
 * Copyright (c) 2004, KTH NADA.
 *
 * This file is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; see the file COPYING.  If not, see
 * http://opensource.org/licenses/gpl-license.php
 *
 * Martin Hassel, 2004-May-20
 * http://www.nada.kth.se/~xmartin/
 */

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.util.HashSet;
import java.util.Set;

import moj.lang.GenericTokenizer;
import moj.lang.StopList;
import moj.lang.en.PorterStemmer;
import moj.lang.se.GranskaConnection;
import moj.ri.SparseDistributedMemory;
import moj.ri.weighting.WeightingScheme;
import moj.util.DirCrawler;
import moj.util.DirCrawlerHelper;
import moj.util.FileTools;
import moj.util.TagStripper;

/**
 * TrainRI indexes a given text, or folder tree of texts, using RandomIndexing
 * given certain properties. In case of <code>OutOfMemoryError</code> or when 
 * being prompted to abort (i.e. Ctrl-c) the index so far is dumped in a well-
 * behaved manner.
 * The settings for connection to a Granska server and indexing are located in 
 * a separate properties files (default values are declared in relevant classes). 
 *
 * @author  Martin Hassel
 * @version 2006-Dec-07
 */
public class TrainRI {
	// Default properties file
	static String propsfile = "TrainRI.properties";
	// Global variables needed for shutdown hook
	static SparseDistributedMemory sdm;
	static String filename;
	static PrintWriter pr;

	/**
	 * *** TrainRI *** <br />
	 * Usage: TrainRI <data> <file to save to> (<properties file>) <br />
	 * <data> : file or folder to recursivly Random Index files in <br />
	 * <file to save to> : filename (without extension) to save the RandomIndex to <br />
	 * <properties file> : properties file to read settings from (optional) <br />
	 */
	public static void main(String arg[]) {
		// Shutdown hook that saves the index in case of OutOfMemoryException
        Thread hook = new Thread() {
    	    public void run() {
    	        System.out.println("Indexing aborted, dumping index so far to compressed file...");
    	        System.out.println("(this behaviour can be altered with the -Xrs JVM option)");
    	    	stopThreads(sdm);
    			sdm.saveCompressed(filename);
    	    }
    	};
        if(arg.length > 1) {
			long time;
			float timediff;
			filename = arg[1];
			if(arg.length > 2)
				propsfile = arg[2];
			// Load and set properties for the RandomIndex
	    	TrainRIprops props = new TrainRIprops(propsfile);
			// Print settings for current index, handy when running batch jobs
	    	props.writeProperties(new OutputStreamWriter(System.out));
			// System.out.println("Added codepage detection...\n");
			try {
				// Create an new RandomIndex and supply attributes and WeightingScheme
	            sdm = new SparseDistributedMemory(props.dimensionality, props.randomDegree, 
	            		props.seed, props.leftWindowSize, props.rightWindowSize, 
	            		(WeightingScheme)Class.forName(props.weightingScheme).newInstance());
				if(!props.startNewIndex) {
					File indexfile0 = new File(filename);
					File indexfile1 = new File(filename+".xml");
					File indexfile2 = new File(filename+".xml.zip");
					if(indexfile0.exists() || indexfile1.exists() || indexfile2.exists()) {
						time = System.currentTimeMillis();
						System.out.print("Loading existing RandomIndex: "+filename);
						if(indexfile0.exists())
							System.out.print(" ("+(indexfile0.length()/1024/1024)+" Mb)");
						else if(indexfile1.exists())
							System.out.print(".xml ("+(indexfile1.length()/1024/1024)+" Mb)");
						else
							System.out.print(".xml.zip ("+(indexfile2.length()/1024/1024)+" Mb)");
						sdm.load(filename);
						timediff = (float)(System.currentTimeMillis()-time)/1000;
						System.out.println(" took " + timediff + " seconds\n");
						System.out.print(sdm.getWordsIndexed() + " words and ");
						System.out.println(sdm.size() + " wordforms/index terms in index");
						if(sdm.isPurged()) {
							System.out.println("\nLabels have been removed from the index, " +
									"further updating might lead to unreliable data\n");
							System.exit(-1);
						}
					}
				} else {
					if(!props.lowercase)
						sdm.setAllLowerCase(false);
					if(props.unaryLabels)
						sdm.setUnaryLabels(props.unaryLabels);
					if(props.documentLabels)
						sdm.setDocumentLabels(props.documentLabels);
				}
				if(props.vocabularyFile.length() > 0) {
					sdm.addVocabulary(readVocabulary(props.vocabularyFile));
			    	System.out.println("Using limited vocabulary, " + sdm.getVocabulary().size() + 
			    			" words loaded from " + props.vocabularyFile + "\n");
				}
				time = System.currentTimeMillis();

				StopList stoplist = new StopList(propsfile);
		    	stoplist.getProperties().writeProperties(new OutputStreamWriter(System.out));
		    	System.out.println(stoplist.getStopWords().size() + " stop words loaded\n");
		    	Runtime.getRuntime().addShutdownHook(hook);

		    	if(props.log)
			    	pr = new PrintWriter(new BufferedWriter(new FileWriter(filename+".log", true)));
	            if(props.numberOfThreads <= 1)
		            sdm.setMaxNumThreads(1);
	            else
	            	sdm.setMaxNumThreads(props.numberOfThreads);
	            System.out.println("number_of_threads = " + sdm.getMaxNumThreads() + "\n");

	            // Allocate instance of helper class that processes each file
		    	TrainRI t = new TrainRI();
		    	Helper helper = t.new Helper(stoplist,props);
				// Now we can recursivly walk through and Random Index the files in the given folder
				DirCrawler.recurseInDirFrom(arg[0], helper);
				timediff = (float)(System.currentTimeMillis()-time)/60000;

				stopThreads(sdm);

		    	if(props.log) {
		            pr.println("Done.");
		            pr.flush();
		            pr.close();
		    	}

				// And print some stats about the index we just built
				System.out.println("");
				System.out.println("Indexing file(s) took " + timediff + " minutes");
				System.out.println(sdm.getWordsIndexed() + " words indexed");
				System.out.println(sdm.size() + " wordforms/index terms");

				// Finally we save the compressed index to the given (path+)filename
				Runtime.getRuntime().removeShutdownHook(hook);
				time = System.currentTimeMillis();
		    	if(props.purge_on_save)
					System.out.println("Removed "+sdm.purgeIndex()+" words not in the limited vocabulary");
				System.out.println("Save compressed: " + props.saveCompressed);
				if(props.saveCompressed)
	    			sdm.saveCompressed(filename);
				else
	    			sdm.save(filename);
				timediff = (float)(System.currentTimeMillis()-time)/1000;
				System.out.println("Saving of RandomIndex took " + timediff + " seconds\n");
			} catch(Exception ex) {
				System.err.println(ex);
	            ex.printStackTrace();
			}
		} else {
			System.out.println("*** TrainRI ***");
			System.out.println("Usage: TrainRI <data> <file to save to> (<properties file>)");
			System.out.println("<data> : file or folder to recursivly Random Index files in");
			System.out.println("<file to save to> : filename (without extension) to save the RandomIndex to");
			System.out.println("<properties file> : properties file to read settings from (optional)");
		}
		Runtime.getRuntime().removeShutdownHook(hook);
	}

	static void stopThreads(SparseDistributedMemory spdm) {
		spdm.finalize();
		// Wait for spawned threads, if any, to finish
		while(!spdm.finishedUpdating()) {
			try {
				System.out.println(Thread.activeCount()+" threads still running");
				Thread.sleep(50);
			} catch(Exception ex) {
				System.err.println(Thread.currentThread().getName()+"Sleep failed");
				ex.printStackTrace();
			}
		}
    	System.out.println("Done.");
	}

	/**
	 * Reads the list of vocabulary words from file and returns a <code>Set</code>. 
	 * @param vocabularyFile filename of the file containg the list words
	 * @return a <code>Set</code> containing the stop words
	 */
	private static Set<String> readVocabulary(String vocabularyFile) {
		Set<String> vocabulary = new HashSet<String>();
		try {
			BufferedReader fr = new BufferedReader(new FileReader(vocabularyFile));
			String word = fr.readLine();
			while(word != null) {
				// Do not add comments
				if(!word.startsWith("#"))
					vocabulary.add(TagStripper.removeControlCodes(word.toLowerCase()));
				word = fr.readLine();
			}
			fr.close();
		} catch(FileNotFoundException fnfe) {
			System.err.println("Warning: Vocabulary file '" + vocabularyFile + "' not found.");
		} catch(IOException ioe) {
			ioe.printStackTrace();
		}
		return vocabulary;
	}

	private class Helper extends DirCrawlerHelper {
		private StopList stoplist;
		private TrainRIprops props;
		
		public Helper(StopList _stoplist, TrainRIprops _props) {
			stoplist = _stoplist;
			props = _props;
		}
		
		public void doForAll(File fileItem) {
			String filename = fileItem.toString();
			if(fileItem.isFile() && filename.endsWith(props.restrictExtension)) {
				System.out.print(filename + " ... ");
				try {
					String text = TagStripper.removeControlCodes(FileTools.readFile2String(filename));
					// Each discourse segment is treated as a separate document
					String[] discourseSegments;
					if(props.discourseMarker.compareToIgnoreCase("")!=0) {
						System.out.println("indexing discourse segments (discourse_marker="+props.discourseMarker+")");
						discourseSegments = text.split(props.discourseMarker);
					} else {
						discourseSegments = new String[1];
						discourseSegments[0] = text;
					}
					for(int i = 0; i < discourseSegments.length; i++) {
						text = discourseSegments[i];
						// Remove HTML tags from the text
						if(props.stripHTML)
							text = new TagStripper().stripHTML(text);
						// Remove stop words from the text
						if(stoplist.getStopWords().size() > 0) {
							text = stoplist.removeStopWords(text, true);
							System.out.print(" ");
						}
						String toktext;
						// Retry n times in case the Granska server fails
						int retries = 0;
						do{
							if(retries > 0)
								System.err.println("Failed to preprocess text - retry " + retries);
							// Granska Server for SWEDISH preprocessing
							if(props.granska.compareToIgnoreCase("tokenize")==0) {
								GranskaConnection granska = new GranskaConnection();
								toktext = granska.tokenize(text);
							} else if(props.granska.compareToIgnoreCase("lemmatize")==0) {
								GranskaConnection granska = new GranskaConnection();
								toktext = granska.lemmatize(text);
							} else if(props.granska.compareToIgnoreCase("lemmatag")==0) {
								GranskaConnection granska = new GranskaConnection();
								toktext = granska.lemmaTag(text);
							} else if(props.granska.compareToIgnoreCase("postag")==0) {
								GranskaConnection granska = new GranskaConnection();
								toktext = granska.posTag(text);
							} else 
								// Porter stemmer for ENGLISH stemming
								if(props.granska.compareToIgnoreCase("porterstemmer")==0) {
								GenericTokenizer gt = new GenericTokenizer();
								PorterStemmer ps = new PorterStemmer();
								StringBuffer sb = new StringBuffer();
								String[] words = gt.getWords(text);
								for(int j = 0; j < words.length; j++)
									sb.append(ps.stripAffixes(words[j])+" ");
								toktext = sb.toString();
							} else {
								// Tokenize text by removing all punctuation
								toktext = text.replaceAll("[\\s+-.,:;!?&(){}<>\"\'/]"," ");
								toktext = toktext.replaceAll("\\s+"," ");
							}
							retries++;
						} while((toktext == null || toktext.length() < 1) && retries < props.retries+1);
						// Something went wrong (e.g. GTA could not parse the text) 
						if(toktext == null) {
							System.err.println("Failed to preprocess text - giving up...");
							if(props.log)
					            pr.println(filename + ": failed");
							toktext = ".";
						} else {
							// Add preprocessed text to Random Index
							int wa = sdm.addText(toktext);
							if(props.log)
					            pr.println(filename + ": " + wa + " words added");
							System.out.println(wa + " words added");
							if(props.savePreprocessedText) {
								// Save preprocessed text to file for later use
								try {
									PrintWriter pri = new PrintWriter(new BufferedWriter(new FileWriter(
											filename + "." + props.granska + "." + i + ".txt", false)));
									pri.print(toktext);
									pri.flush();
									pri.close();
								} catch(java.io.IOException ex) {
									System.err.println(ex);
									ex.printStackTrace();
								}
							}
						}
						if(props.log)
				            pr.flush();
					}
				} catch(Exception ex) {
					System.err.println(ex);
		            ex.printStackTrace();
				}
			}
		}
	}
}
