/*
 * FrequencyList.java
 * Copyright (c) 2004, KTH NADA.
 *
 * This file is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; see the file COPYING.  If not, see
 * http://opensource.org/licenses/gpl-license.php
 *
 * Martin Hassel, 2004-Oct-08
 * http://www.nada.kth.se/~xmartin/
 */

package moj.lang;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import moj.util.FileTools;

/**
 * <code>FrequencyList</code> holds a list of tokens (index terms) and their 
 * respective term and document frequencies in the texts indexed (or 'added') 
 * so far. Sublists can be extracted based on frequency thresholds.
 *
 * @author  Martin Hassel
 * @version 2006-Oct-30
 */
public class FrequencyList {
	private Map<String,int[]> _frequencyList = new HashMap<String,int[]>();
	private int _wordCount = 0;
	private int _documentCount = 0;
	private int _highestTF = 0;
	private int _lowestTF = Integer.MAX_VALUE;
	private int _highestDF = 0;
	private int _lowestDF = Integer.MAX_VALUE;

	/**
	 * Creates a new empty <code>FrequencyList</code>.
	 */
	public FrequencyList() {}

	/**
	 * Creates a new <code>FrequencyList</code> based upon a <code>Map</code> 
	 * containing <code>String</code>s as keys, and values consisting of an 
	 * <code>int</code> array of the size 2 where the first element is the 
	 * term frequency and the second is the document frequency.
	 * 
	 * @param frequencyList <code>Map</code> containing <code>String</code>s 
	 *        as keys, and values consisting of an <code>int</code> array of the 
	 *        size 2 where the first element is the term frequency and the 
	 *        second is the document frequency
	 */
	private FrequencyList(Map<String,int[]> frequencyList, int wordCount, int documentCount, 
			int highestTF, int lowestTF, int highestDF, int lowestDF) {
		_frequencyList = frequencyList;
		_wordCount = wordCount;
		_documentCount = documentCount;
		_highestTF = highestTF;
		_lowestTF = lowestTF;
		_highestDF = highestDF;
		_lowestDF = lowestDF;
	}

	/**
	 * Extract sublist containing only the index terms with a term frequency 
	 * between <code>lowerBound</code> <i>and</i> <code>upperBound</code>.
	 * 
	 * @param lowerBound lowest acceptable term frequency
	 * @param upperBound highest acceptable term frequency
	 * @return a new <code>FrequencyList</code> containing the sublist
	 */
	public FrequencyList filterByTFinside(int lowerBound, int upperBound) {
		return filterByFrequencyInside(lowerBound, upperBound, 0);
	}

	/**
	 * Extract sublist containing only the index terms with a document frequency 
	 * between <code>lowerBound</code> <i>and</i> <code>upperBound</code>.
	 * 
	 * @param lowerBound lowest acceptable document frequency
	 * @param upperBound highest acceptable document frequency
	 * @return a new <code>FrequencyList</code> containing the sublist
	 */
	public FrequencyList filterByDFinside(int lowerBound, int upperBound) {
		return filterByFrequencyInside(lowerBound, upperBound, 1);
	}

	/**
	 * Extract sublist containing only the index terms with a term or document 
	 * frequency between <code>lowerBound</code> <i>and</i> <code>upperBound</code>.
	 * 
	 * @param lowerBound lowest acceptable term or document frequency
	 * @param upperBound highest acceptable term or document frequency
	 * @param field 0=tf and 1=df
	 * @return a new <code>FrequencyList</code> containing the sublist
	 */
	private FrequencyList filterByFrequencyInside(int lowerBound, int upperBound, int field) {
        int wordCount = 0, highestTF = 0, lowestTF = Integer.MAX_VALUE;
		int highestDF = 0, lowestDF = Integer.MAX_VALUE;
		Map<String,int[]> frequencyList = new HashMap<String,int[]>();
		for(Iterator iterator = _frequencyList.keySet().iterator(); iterator.hasNext();) {
        	String token = (String)iterator.next();
			int[] tfdf = _frequencyList.get(token);
        	if(tfdf[field] > lowerBound && tfdf[field] < upperBound) {
        		frequencyList.put(token, tfdf);
	        	if(tfdf[0] > highestTF)
	        		highestTF = tfdf[0];
	        	if(tfdf[0] < lowestTF)
	        		lowestTF = tfdf[0];
	        	if(tfdf[1] > highestDF)
	        		highestDF = tfdf[1];
	        	if(tfdf[1] < lowestDF)
	        		lowestDF = tfdf[1];
	        	wordCount += tfdf[0];
        	}
        }
        return new FrequencyList(frequencyList, wordCount, _documentCount, 
				highestTF, lowestTF, highestDF, lowestDF);
	}

	/**
	 * Extract sublist containing only the index terms with a term frequency 
	 * lower than <code>lowerBound</code> <i>or</i> higher than <code>upperBound</code>.
	 * 
	 * @param lowerBound lower term frequency boundary
	 * @param upperBound higher term frequency boundary
	 * @return a new <code>FrequencyList</code> containing the sublist
	 */
	public FrequencyList filterByTFoutside(int lowerBound, int upperBound) {
		return filterByFrequencyOutside(lowerBound, upperBound, 0);
	}

	/**
	 * Extract sublist containing only the index terms with a document frequency 
	 * lower than <code>lowerBound</code> <i>or</i> higher than <code>upperBound</code>.
	 * 
	 * @param lowerBound lower document frequency boundary
	 * @param upperBound higher document frequency boundary
	 * @return a new <code>FrequencyList</code> containing the sublist
	 */
	public FrequencyList filterByDFoutside(int lowerBound, int upperBound) {
		return filterByFrequencyOutside(lowerBound, upperBound, 1);
	}

	/**
	 * Extract sublist containing only the index terms with a term or document 
	 * frequency lower than <code>lowerBound</code> <i>or</i> higher than 
	 * <code>upperBound</code>.
	 * 
	 * @param lowerBound lower term or document frequency boundary
	 * @param upperBound higher term or document frequency boundary
	 * @param field 0=tf and 1=df
	 * @return a new <code>FrequencyList</code> containing the sublist
	 */
	private FrequencyList filterByFrequencyOutside(int lowerBound, int upperBound, int field) {
        int wordCount = 0, highestTF = 0, lowestTF = Integer.MAX_VALUE;
		int highestDF = 0, lowestDF = Integer.MAX_VALUE;
		Map<String,int[]> frequencyList = new HashMap<String,int[]>();
		for(Iterator iterator = _frequencyList.keySet().iterator(); iterator.hasNext();) {
        	String token = (String)iterator.next();
			int[] tfdf = _frequencyList.get(token);
        	if(tfdf[field] < lowerBound || tfdf[field] > upperBound) {
        		frequencyList.put(token, tfdf);
	        	if(tfdf[0] > highestTF)
	        		highestTF = tfdf[0];
	        	if(tfdf[0] < lowestTF)
	        		lowestTF = tfdf[0];
	        	if(tfdf[1] > highestDF)
	        		highestDF = tfdf[1];
	        	if(tfdf[1] < lowestDF)
	        		lowestDF = tfdf[1];
	        	wordCount += tfdf[0];
        	}
        }
        return new FrequencyList(frequencyList, wordCount, _documentCount, 
				highestTF, lowestTF, highestDF, lowestDF);
	}

	/**
	 * @return the highest term frequency in the frequency list
	 */
	public int getHighestTF() {
		return _highestTF;
	}

	/**
	 * @return the lowest term frequency in the frequency list
	 */
	public int getLowestTF() {
		return _lowestTF;
	}

	/**
	 * @return the highest document frequency in the frequency list
	 */
	public int getHighestDF() {
		return _highestDF;
	}

	/**
	 * @return the lowest document frequency in the frequency list
	 */
	public int getLowestDF() {
		return _lowestDF;
	}

	/**
	 * @return the number of documents represented in the frequency list
	 */
	public int getDocumentCount() {
		return _documentCount;
	}

	/**
	 * @return the number of words represented in the frequency list
	 */
	public int getWordCount() {
		return _wordCount;
	}

	/**
	 * This gives the same result as <code>getUniqueWords().size()</code>
	 * @return the number of unique words represented in the frequency list
	 */
	@Deprecated
	public int getUniqueWordCount() {
		return _frequencyList.size();
	}

	/**
	 * @return a <code>Set</code> containing each unique word represented 
	 *         in the frequency list
	 */
	public Set<String> getUniqueWords() {
		return _frequencyList.keySet();
	}

	/**
	 * Returns the term frequency for the given word. If the word is not 
	 * in the frequency list then a frequency of zero is returned.
	 * 
	 * @param word the word (token) to get the term frequency for
	 * @return the term frequency for the given word
	 */
	public int getTF(String word) {
		if(_frequencyList.containsKey(word))
			return _frequencyList.get(word)[0];
		return 0;
	}

	/**
	 * Returns the document frequency for the given word. If the word is not 
	 * in the frequency list then a frequency of zero is returned.
	 * 
	 * @param word the word (token) to get the document frequency for
	 * @return the document frequency for the given word
	 */
	public int getDF(String word) {
		if(_frequencyList.containsKey(word))
			return _frequencyList.get(word)[1];
		return 0;
	}

	/**
	 * Adds tokenized text to the <code>FrequencyList</code> and updates the 
	 * term and document frequencies for each encountered token.
	 * 
	 * @param text tokenized text to be added
	 * @return number of tokens (words) added to <code>FrequencyList</code>
	 */
	public int addText(String[] text) {
		return addText(text, 0);
	}

	/**
	 * Adds tokenized text to the <code>FrequencyList</code> and updates the 
	 * term and document frequencies for each encountered token. Tokens with 
	 * a length (in characters) shorter than <code>mintokenlen</code> will not 
	 * be added to the list.
	 * 
	 * @param text tokenized text to be added
	 * @param mintokenlen minimum length of a token for it to be added to the list
	 * @return number of tokens (words) added to <code>FrequencyList</code>
	 */
	public int addText(String[] text, int mintokenlen) {
		int added = 0;
        Set<String> uniqueWords = new HashSet<String>();
        for(int current = 0; current < text.length; current++) {
            String token = text[current].toLowerCase();
            if(token.length() >= mintokenlen) {
	            if(_frequencyList.get(token) == null) {
	            	// Create new entry with a tf of 1.
	            	int[] tfdf = {1,0};
	            	_frequencyList.put(token, tfdf);
	            } else {
	            	// Increment existing tf by 1.
					int[] tfdf = _frequencyList.get(token);
	            	int[] newtf = {tfdf[0]+1, tfdf[1]};
	            	_frequencyList.put(token, newtf);
	            }
	            uniqueWords.add(token);
	            _wordCount++;
				added++;
            }
        }
    	// Increment existing df by 1 for all unique words in this text
		for(Iterator iterator = uniqueWords.iterator(); iterator.hasNext();) {
            String token = (String)iterator.next();
			int[] tfdf = _frequencyList.get(token);
			int tf = tfdf[0];
			int df = tfdf[1]+1;
        	if(tf > _highestTF)
        		_highestTF = tf;
        	if(tf < _lowestTF)
        		_lowestTF = tf;
        	if(df > _highestDF)
        		_highestDF = df;
        	if(df < _lowestDF)
        		_lowestDF = df;
        	int[] newdf = {tf, df};
        	_frequencyList.put(token, newdf);
		}
        _documentCount++;
        return added;
    }

    /**
     * Stores index terms and their corresponding term and document frequencies 
     * in the parallel array parameters. The arrays will be filled with values 
     * until full, or with at most <code>getUniqueWords().size()</code> values.
     * The index terms are sorted alphabetically.
     * 
     * @param words array to be filled with the unique words that were encountered 
     *        in the source document(s)
     * @param tf array to be filled with the term frequencies of the words at 
     *        corresponding index in the words array
     * @param df array to be filled with the document frequencies of the words at 
     *        corresponding index in the words array
     */
    public void getWordFrequency(String[] words, int[] tf, int[] df) {
    	getWordFrequency(words, tf, df, new CompareAlphabetically());
    }

    /**
     * Stores index terms and their corresponding term and document frequencies 
     * in the parallel array parameters. The arrays will be filled with values 
     * until full, or with at most <code>getUniqueWords().size()</code> values.
     * The index terms are sorted according to the comparator <code>comp</code>.
     * 
     * @param words array to be filled with the unique words that were encountered 
     *        in the source document(s)
     * @param tf array to be filled with the term frequencies of the words at 
     *        corresponding index in the words array
     * @param df array to be filled with the document frequencies of the words at 
     *        corresponding index in the words array
     * @param comp <code>Comparator</code> denoting how the parallell arrays 
     *        should be ordered
     */
    public void getWordFrequency(String[] words, int[] tf, int[] df, Comparator<Object> comp) {
        List<Map.Entry> entries = new ArrayList<Map.Entry>(_frequencyList.entrySet());
        Collections.sort(entries, comp);
        for(int i = 0; i < entries.size(); i++) {
            Map.Entry entry = entries.get(i);
            if(i < words.length)
                words[i]  = (String)entry.getKey();
            if(i < tf.length)
                tf[i] = ((int[])entry.getValue())[0];
            if(i < df.length)
                df[i] = ((int[])entry.getValue())[1];
        }
    }

    public String toString() {
    	return toString(new CompareAlphabetically());
    }

    public String toString(Comparator<Object> comp) {
		StringBuffer sbuf = new StringBuffer();
		List<Map.Entry> entries = new ArrayList<Map.Entry>(_frequencyList.entrySet());
		Collections.sort(entries, comp);
        for(int i = 0; i < entries.size(); i++) {
            Map.Entry entry = entries.get(i);
        	if(((String)entry.getKey()).length() > 0)
        	sbuf.append(entry.getKey() + "\t" + 
        			((int[])entry.getValue())[0] + "\t" + 
        			((int[])entry.getValue())[1] + "\n");
        }
        return sbuf.toString();
    }

    /**
     * For ordering words alphabetically.
     */
    public class CompareAlphabetically implements Comparator<Object> {
        public int compare(Object o1, Object o2) {
        	return ((String)((Map.Entry)o1).getKey()).compareTo((String)((Map.Entry)o2).getKey());
    	}
    }

    /**
     * For ordering words from lowest to highest term frequency.
     * If frequency is equal, words are put in alphabetical order.
     */
    public class CompareByTF implements Comparator<Object> {
        public int compare(Object o1, Object o2) {
            return compare((Map.Entry)o1, (Map.Entry)o2);
        }
        public int compare(Map.Entry obj1, Map.Entry obj2) {
            int c1 = ((int[])obj1.getValue())[0];
            int c2 = ((int[])obj2.getValue())[0];
            if (c1 < c2) {
                return -1;
            } else if (c1 > c2) {
                return 1;
            } else { // If counts are equal, compare keys alphabetically.
                return ((String)obj1.getKey()).compareTo((String)obj2.getKey());
            }
        }
    }

    /**
     * For ordering words from lowest to highest term frequency.
     * If frequency is equal, words are put in alphabetical order.
     */
    public class CompareByTFfalling implements Comparator<Object> {
        public int compare(Object o1, Object o2) {
            return compare((Map.Entry)o1, (Map.Entry)o2);
        }
        public int compare(Map.Entry obj1, Map.Entry obj2) {
            int c1 = ((int[])obj1.getValue())[0];
            int c2 = ((int[])obj2.getValue())[0];
            if (c1 > c2) {
                return -1;
            } else if (c1 < c2) {
                return 1;
            } else { // If counts are equal, compare keys alphabetically.
                return ((String)obj1.getKey()).compareTo((String)obj2.getKey());
            }
        }
    }

    /**
     * For ordering words from lowest to highest document frequency.
     * If frequency is equal, words are put in alphabetical order.
     */
    public class CompareByDF implements Comparator<Object> {
        public int compare(Object o1, Object o2) {
            return compare((Map.Entry)o1, (Map.Entry)o2);
        }
        public int compare(Map.Entry obj1, Map.Entry obj2) {
            int c1 = ((int[])obj1.getValue())[1];
            int c2 = ((int[])obj2.getValue())[1];
            if (c1 < c2) {
                return -1;
            } else if (c1 > c2) {
                return 1;
            } else { // If counts are equal, compare keys alphabetically.
                return ((String)obj1.getKey()).compareTo((String)obj2.getKey());
            }
        }
    }

    /**
     * For ordering words from highest to lowest document frequency.
     * If frequency is equal, words are put in alphabetical order.
     */
    public class CompareByDFfalling implements Comparator<Object> {
        public int compare(Object o1, Object o2) {
            return compare((Map.Entry)o1, (Map.Entry)o2);
        }
        public int compare(Map.Entry obj1, Map.Entry obj2) {
            int c1 = ((int[])obj1.getValue())[1];
            int c2 = ((int[])obj2.getValue())[1];
            if (c1 > c2) {
                return -1;
            } else if (c1 < c2) {
                return 1;
            } else { // If counts are equal, compare keys alphabetically.
                return ((String)obj1.getKey()).compareTo((String)obj2.getKey());
            }
        }
    }

	/**
	 * Usage: moj.lang.FrequencyList &lt;file&gt; (&lt;minimum token length&gt;)<br />
	 * &lt;file&gt; : file to build frequency counts on<br />
	 * &lt;minimum token length&gt; : minimum length for a token for it to be counted<br />
	 */
	public static void main(String[] args) {
        if(args.length == 0) {
    		System.out.println("*** FrequencyList ***");
    		System.out.println("Usage: moj.lang.FrequencyList <file> (<minimum token length>)");
    		System.out.println("<file> : file to build frequency counts on");
    		System.out.println("<minimum token length> : minimum length for a token for it to be counted");
    		System.exit(0);
    	}
        int mintokenlen = 0;
        if(args.length > 1)
        	mintokenlen = Integer.parseInt(args[1]);
        String text = FileTools.readFile2String(args[0]);
		FrequencyList fl = new FrequencyList();
		GenericTokenizer gt = new GenericTokenizer();
		fl.addText(gt.getWords(text), mintokenlen);
    	System.out.println(fl.toString(fl.new CompareByTF()));
	}
}
