/*
 * GenericTokenizer.java
 * Copyright (c) 2004, KTH NADA.
 *
 * This file is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; see the file COPYING.  If not, see
 * http://opensource.org/licenses/gpl-license.php
 *
 * Martin Hassel, 2004-Nov-17
 * http://www.nada.kth.se/~xmartin/
 */

package moj.lang;

import java.text.BreakIterator;
import java.util.Iterator;
import java.util.Locale;
import java.util.Set;
import java.util.TreeSet;

import moj.util.ArrayTools;
import moj.util.FileTools;

/**
 * GenericTokenizer is a wrapper class that tries to extend <code>BreakIterator</code>s
 * usability. It does so by removing token boundaries set by <code>BreakIterator</code> 
 * where an abbreviation from a supplied abbreviation list is encountered.
 * It also provides handy methods for converting a <code>String</code> of text 
 * to an array of <code>String</code> tokens, which can be either the characters, 
 * words, sentences or the vocabulary of the text.
 *
 * @author  Martin Hassel
 * @version 2006-Oct-30
 */
public class GenericTokenizer {
	private final Set<String> _abbreviations = new TreeSet<String>();
	private Locale _where = Locale.ENGLISH;

	/**
	 * Create a new <code>GenericTokenizer</code> with English locale.
	 */
	public GenericTokenizer() {}

	/**
	 * Create a new <code>GenericTokenizer</code> with the supplied locale.
	 * 
	 * @param where which language settings we want to use.
	 */
	public GenericTokenizer(Locale where) {
		_where = where;
	}

	/**
	 * Get the <code>Locale</code> to which the tokenizer should conform to.
	 * 
	 * @return the currently set <code>Locale</code>
	 */
	public Locale getLocale() {
		return _where;
	}

	/**
	 * Set the <code>Locale</code> to which the tokenizer should conform to.
	 * 
	 * @param where <code>Locale</code> to be set
	 * @return the previously set <code>Locale</code>
	 */
	public Locale setLocale(Locale where) {
		Locale old = _where;
		_where = where;
		return old;
	}

	/**
	 * Wrap the supplied <code>String</code> at a certain width by inserting 
	 * line breaks into the text at proper places without breaking non-hyphenated 
	 * words. The formatted string is returned in a <code>String</code> array 
	 * where each element is a line.
	 * 
	 * @param text the text we want to have wrapped at a cretain width
	 * @param width the maximum width when breaking the text into lines
	 * @return an array with all new lines
	 */
	public String[] wrapText(String text, int width) {
		StringBuffer tempbuf = new StringBuffer();
		BreakIterator boundary = BreakIterator.getLineInstance(_where);
		text = text.replaceAll(System.getProperty("line.separator")," ").replaceAll("\\s+"," ");
        boundary.setText(text);
        int start = boundary.first(), lastbreak = 0;
	    for(int end = boundary.next(); end != BreakIterator.DONE;) {
	    	String token = text.substring(start, end);
    		start = end;
    		end = boundary.next();
    		if(token != "\\s+" && token != "") {
    			if(end - lastbreak < width)
            		tempbuf.append(token);
    			else {
            		tempbuf.append("<tbr/>" + token);
            		lastbreak = end;
    			}
    		}
	    }
	    return tempbuf.toString().split("<tbr/>");
	}

	/**
	 * Tokenize the supplied <code>String</code> into character tokens and 
	 * return them in a <code>String</code> array.
	 * 
	 * @param text the text we want to have tokenized into characters
	 * @return an array with all characters in <code>text</code>
	 */
	public String[] getCharacters(String text) {
		return markTokens(BreakIterator.getCharacterInstance(_where), text).split("<tbr/>");
	}

	/**
	 * Tokenize the supplied <code>String</code> into words and punctuation 
	 * tokens and return them in a <code>String</code> array.
	 * 
	 * @param text the text we want to have tokenized into words
	 * @return an array with all words in <code>text</code>
	 */
	public String[] getWords(String text) {
		return getWords(text, false);
	}

	/**
	 * Tokenize the supplied <code>String</code> into word level tokens and 
	 * return them in a <code>String</code> array.
	 * 
	 * @param text the text we want to have tokenized into word level tokens
	 * @param removePunctuation removes punctuation before tokenizing if 
	 *        <code>true</code>
	 * @return an array with all words in <code>text</code>
	 */
	public String[] getWords(String text, boolean removePunctuation) {
		if(removePunctuation)
			text = text.replaceAll("[\\s\\[\\](){}<>/+-.,:;!?&\"\']"," ");
		text = markTokens(BreakIterator.getWordInstance(_where), text);
		text = text.replaceAll("\\s", "");
		text = text.replaceAll("(<tbr/>)+", "<tbr/>");
		text = text.replaceAll("^<tbr/>", "").replaceAll("<tbr/>$", "");
		return text.split("<tbr/>");
	}

	/**
	 * Tokenize the supplied <code>String</code> into sentence tokens and return 
	 * them in a <code>String</code> array.
	 * 
	 * @param text the text we want to have tokenized into sentences
	 * @return an array with all sentences in <code>text</code>
	 */
	public String[] getSentences(String text) {
		text = markTokens(BreakIterator.getSentenceInstance(_where), text);
		text = text.replaceAll("^\\s+", "").replaceAll("\\s+$", "");
		for(Iterator iterator = _abbreviations.iterator(); iterator.hasNext();) {
	    	String abbreviation = (String)iterator.next();
	    	if(abbreviation != null && abbreviation.length() > 0)
	    		text = text.replaceAll("(\\s|<tbr/>)"+abbreviation+"\\s<tbr/>", "$1"+abbreviation+" ");
	    }
		return text.split("\\s<tbr/>");
	}

	/**
	 * Mark token boundaries.
	 * 
	 * @param boundary boundary type
	 * @param text text to mark token boundaries in
	 * @return <code>String text</code> with token boundaries marked
	 */
	private String markTokens(BreakIterator boundary, String text) {
		StringBuffer tempbuf = new StringBuffer();
		text = text.replaceAll(System.getProperty("line.separator")," ").replaceAll("\\s+"," ");
        boundary.setText(text);
        int start = boundary.first();
	    for(int end = boundary.next(); end != BreakIterator.DONE;) {
	    	String token = text.substring(start, end);
    		start = end;
    		end = boundary.next();
    		if(token != "\\s+" && token != "") {
        		tempbuf.append(token);
        		if(end != BreakIterator.DONE)
            		tempbuf.append("<tbr/>");
    		}
	    }
	    return tempbuf.toString();
	}

	/**
	 * Extract the vocabulary, i.e. all unique word tokens from the supplied 
	 * <code>String</code> and return them (alphabetically) sorted 
	 * in an array. Punctuation tokens are removed.
	 * 
	 * @param text the text we want the vocabulary of
	 * @return an array with all unique words in <code>text</code>
	 */
	public String[] getVocabulary(String text) {
		return getVocabulary(text, true);
	}

	/**
	 * Extract the vocabulary, i.e. all unique tokens ("words"), from the 
	 * supplied <code>String</code> and return them (alphabetically) sorted 
	 * in an array.
	 * 
	 * @param text the text we want the vocabulary of
	 * @param removePunctuation removes punctuation before tokenizing if 
	 *        <code>true</code>
	 * @return an array with all unique tokens ("words") in <code>text</code>
	 */
	public String[] getVocabulary(String text, boolean removePunctuation) {
		Set<Object> vocabulary = new TreeSet<Object>();
		ArrayTools.addAll(vocabulary, getWords(text,removePunctuation));
	    return vocabulary.toArray(new String[0]);
	}

	/**
	 * Add abbreviation to the tokenizers internal list of abbreviations.
	 * 
	 * @param abbreviation abbreviation to add
	 */
	public void addAbbreviation(String abbreviation) {
		_abbreviations.add(abbreviation);
	}
	
	/**
	 * Add a set of abbreviations from an array containing exactly one abbreviation 
	 * per element.
	 * 
	 * @param abbreviations abbreviations to be added to the tokenizer
	 */
	public void addAbbreviations(String[] abbreviations) {
		for(int i = 0; i < abbreviations.length; i++)
			_abbreviations.add(abbreviations[i]);
	}

	/**
	 * Add a set of abbreviations from a file containing exactly one abbreviation 
	 * per line.
	 * 
	 * @param abbreviationFile full path to file containing abbreviationss to 
	 *        be added to the tokenizer
	 */
	public int addAbbreviations(String abbreviationFile) {
		int added = 0;
		String[] abbreviations = FileTools.readFile2String(abbreviationFile).split(System.getProperty("line.separator"));
		for(int i = 0; i < abbreviations.length; i++) {
			String abbreviation = abbreviations[i];
			// Do not add comments
			if(!abbreviation.startsWith("#")) {
				_abbreviations.add(abbreviation);
				added++;
			}
		}
		return added;
	}

	/**
	 * Usage: moj.lang.GenericTokenizer &lt;file&gt; (&lt;token type&gt;)<br />
	 * &lt;file&gt; : file to tokenize<br />
	 * &lt;token type&gt; : ([s]entenences|[w]ords|[v]ocabulary|[c]haracters|[l]inebreak)<br />
	 */
	public static void main(String[] args) {
        if(args.length == 0) {
    		System.out.println("*** GenericTokenizer ***");
    		System.out.println("Usage: moj.lang.GenericTokenizer <file> (<token type>)");
    		System.out.println("<file> : file to build frequency counts on");
    		System.out.println("<token type> : ([s]entenences|[w]ords|[v]ocabulary)|[c]haracters|[l]inebreak");
    		System.exit(0);
    	}
		String[] tokens = {};
		GenericTokenizer gt = new GenericTokenizer();
		gt.addAbbreviations("abbreviations.txt");
		String text = FileTools.readFile2String(args[0]);
		//text = "Yesterday I met Mr. and Mrs. Smith. They were on their way to see Dr. Jekyll vs. Prof. Hyde, which is supposed to be such a good movie. A film about good vs. evil is always a good treat.";
        if(args.length > 1 && (args[1].startsWith("w")))
			tokens = gt.getWords(text, true);
        else if(args.length > 1 && (args[1].startsWith("c")))
			tokens = gt.getCharacters(text);
        else if(args.length > 1 && (args[1].startsWith("v")))
			tokens = gt.getVocabulary(text);
        else if(args.length > 1 && (args[1].startsWith("l")))
			tokens = gt.wrapText(text, 80);
        else
        	tokens = gt.getSentences(text);
		for(int token = 0; token < tokens.length; token++)
	    	System.out.println(tokens[token]);
	}
}
