/*
 */
package moj.lang;

import moj.util.FileTools;

/**
 *
 * @author  Martin Hassel
 * @version 2006-Nov-06
 */
public class Readability {

	/**
	 * Calculate the LIX value for a given <code>String</code> of text.<br />
	 * <br />
	 * LIX is a Swedish acronym for "L&auml;sbarhetsindex", or Readability Index, and is 
	 * a much disputed measure of how accessible a text is. The higher the LIX the 
	 * harder, arguably, the text is to understand. For a short description of LIX 
	 * in Swedish, as well as a table of how to interpret the LIX value according to 
	 * prespecified "readability classes", please see: http://sv.wikipedia.org/wiki/LIX
	 * 
	 * @param text <code>String</code> to calculate LIX for
	 * @return the LIX value for the given <code>String</code>
	 */
	public static float calcLIX(String text) {
		GenericTokenizer gt = new GenericTokenizer();
		gt.addAbbreviations("abbreviations.txt");
       	return calcLIX(text, gt);
	}

	/**
	 * Calculate the LIX value for a given <code>String</code> of text using 
	 * the provided <code>GenericTokenizer</code> for splitting the text into 
	 * sentences and words.
	 * 
	 * @param text <code>String</code> to calculate LIX for
	 * @param gt <code>GenericTokenizer</code> to use for tokenizing the text
	 * @return the LIX value for the given <code>String</code>
	 */
	public static float calcLIX(String text, GenericTokenizer gt) {
		float longwords = 0;
		String[] words = gt.getWords(text, true);
       	for(int current = 0; current < words.length; current++)
			if(words[current].length() > 6)
				longwords++;
       	return ((longwords*100)/words.length) + ((float)words.length/(float)gt.getSentences(text).length);
	}

	public static float calcSMOG(String text) {
		GenericTokenizer gt = new GenericTokenizer();
		gt.addAbbreviations("abbreviations.txt");
       	return calcLIX(text, gt);
	}

	public static float calcSMOG(String text, GenericTokenizer gt) {
		float polysyllables = 0;
		String[] words = gt.getWords(text, true);
       	for(int current = 0; current < words.length; current++)
			if(calcSyllables(words[current]) > 3)
				polysyllables++;
		return (float)(1.0430*Math.sqrt(polysyllables*(30/(float)gt.getSentences(text).length))+3.1291);
	}

	/**
	 * Note that this is not infallible.  it does fail for some percentage of 
	 * words (10% seems a good guess)...  so it's useful for approximation, but 
	 * don't use this for running your nuclear reactor...<br />
	 * &copy; Greg Fast (gdf@imsa.edu)
	 */
	public static int calcSyllables(String word) {
		//$VERSION = '0.251';
		//$REVISION = '$Id: Readability.java,v 1.3 2006/12/06 13:54:54 xmartin Exp $ ';
		// basic algortithm:
		// each vowel-group indicates a syllable, except for:
		//  final (silent) e
		//  'ia' ind two syl 
		// @AddSyl and @SubSyl list regexps to massage the basic count.
		// Each match from @AddSyl adds 1 to the basic count, each @SubSyl match -1
		// Keep in mind that when the regexps are checked, any final 'e' will have
		// been removed, and all '\'' will have been removed.
		String[] subsyl = {
				"cial",
				"tia",
				"cius",
				"cious",
				"giu",              // belgium!
				"ion",
				"iou",
				"sia$",
				".ely$",             // absolutely! (but not ely!)
		};
		String[] addsyl = { 
				"ia",
				"riet",
				"dien",
				"iu",
				"io",
				"ii",
				"[aeiouym]bl$",     // -Vble, plus -mble
				"[aeiou]{3}",       // agreeable
				"^mc",
				"ism$",             // -isms
				"([^aeiouy])\1l$",  // middle twiddle battle bottle, etc.
				"[^l]lien",         // alien, salient [1]
				"^coa[dglx].",      // [2]
				"[^gq]ua[^auieo]",  // i think this fixes more than it breaks
				"dnt$",           // couldn't
		};
		
		word = word.toLowerCase();
		word = word.replace("'","");
		word = word.endsWith("e") ? word.substring(0,word.length()-1) : word;
		String[] scrugg = word.split("[^aeiouy]+"); // '-' should perhaps be added?
		if(scrugg.length != 0 && scrugg[0] == "")
			System.arraycopy(scrugg, 1, scrugg, 0, scrugg.length-1);
		int sylcount = 0;
		// special cases
		for(String syl : subsyl)
			if(word.indexOf(syl) > 0)
				sylcount--;
		for(String syl : addsyl)
			if(word.indexOf(syl) > 0)
				sylcount++;
		if(word.length() == 1)
			sylcount++;	// 'x'
		// count vowel groupings
		sylcount += scrugg.length;
		sylcount = sylcount==0 ? 1 : sylcount; // got no vowels? ("the", "crwth")
		return sylcount;
	}

	/**
	 * Usage: moj.lang.Readability &lt;file&gt;<br />
	 * &lt;file&gt; : text file to calculate LIX for<br />
	 */
	public static void main(String[] args) {
        if(args.length == 0) {
    		System.out.println("*** Readability ***");
    		System.out.println("Usage: moj.lang.Readability <file>");
    		System.out.println("<file> : text file to calculate readability for");
    		System.exit(0);
    	}
		String text = FileTools.readFile2String(args[0]);
		GenericTokenizer gt = new GenericTokenizer();
		gt.addAbbreviations("abbreviations.txt");
		String[] words = gt.getWords(text, true);
       	float nrofwords = words.length;
       	System.out.println("Number of words: " + nrofwords);
       	float nrlongwords = 0;
       	for(int current = 0; current < words.length; current++)
			if(words[current].length() > 6)
				nrlongwords++;
       	System.out.println("Number of long words: " + nrlongwords);
       	float longwordwordratio = (nrlongwords*100)/nrofwords;
       	System.out.println("Long words / all words: " + longwordwordratio);
       	float nrofsents = gt.getSentences(text).length;
       	System.out.println("Number of sentences: " + nrofsents);
       	float longwordsentratio = nrofwords/nrofsents;
       	System.out.println("Words / sentences: " + longwordsentratio);
       	float LIX = longwordwordratio + longwordsentratio;
       	System.out.println("LIX for this text: " + LIX);
       	System.out.println("SMOG for this text: " + calcSMOG("How readable is the writing in your publication or on your website? Here's a handy tool for estimating the reading level of a book, article, or web page, based on the SMOG readability formula. This formula was first published by Dr. G. Harry McLaughlin in 1969, and it has been widely used by editors and writers ever since. Recently, McLaughlin teamed up with JAVA programming expert Alain Trottier to develop an online SMOG calculator that can analyze blocks of text up to 2,000 words long, easily and instantly. It's a valuable tool for editors, writers, and anyone who cares about readable writing. You can give their calculator a try here:",gt));
	}
}
