/*
 * StandardDeviationList.java
 * Copyright (c) 2005, KTH NADA.
 *
 * This file is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; see the file COPYING.  If not, see
 * http://opensource.org/licenses/gpl-license.php
 *
 * Martin Hassel, 2005-Oct-19
 * http://www.nada.kth.se/~xmartin/
 */

package moj.lang;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import moj.util.EntryValueComparator;
import moj.util.FileTools;

/**
 * <code>StandardDeviationList</code> holds a list of unique tokens (index terms) 
 * and their respective standard deviation of the distance between the token 
 * instances' positions in the text processed by the list. This can e.g. be used 
 * for keyword extraction as a high standard deviation indicates a high "burstiness" 
 * for the token's distribution in the text (i.e. a good keyword indicator for 
 * longer texts).
 * For more information see "Keyword detection in natural languages and DNA" 
 * (Ortuno et al 2002).
 *
 * @author  Martin Hassel
 * @version 2006-Oct-23
 */
public class StandardDeviationList {
	private HashMap<String,Double> _standardDeviations = new HashMap<String,Double>();
	private HashMap<String,ArrayList<Integer>> _distances = new HashMap<String,ArrayList<Integer>>();
	private HashMap<String,Integer> _totalDistance = new HashMap<String,Integer>();
	private HashMap<String,Integer> _lastPosition;

	/**
	 * Creates a new empty <code>StandardDeviationList</code>.
	 */
	public StandardDeviationList() {}

	/**
	 * Gets a <code>HashMap</code> containing the standard deviations of the 
	 * distances between each instance of each unique token in the added text(s).
	 * 
	 * @return <code>HashMap</code> containing standard deviations of token distances
	 */
	public HashMap<String,Double> getStandardDeviations() {
		return _standardDeviations;
	}

	/**
	 * Adds tokenized text to the <code>StandardDeviationList</code> 
	 * and takes notes of every tokens position(s) in the text.
	 * 
	 * @param text tokenized text to be added
	 */
	public int addText(String[] text) {
		return addText(text, 0);
	}

	/**
	 * Adds tokenized text to the <code>StandardDeviationList</code> 
	 * and takes notes of every tokens position(s) in the text.
	 * 
	 * @param text tokenized text to be added
	 * @param mintokenlen minimum length of a token for it to be added to the list
	 */
	public int addText(String[] text, int mintokenlen) {
		int counted = 0;
		// Calculate and store the distance between instances of each unique token 
		// as well as the total amounted distance between the token instances
		_lastPosition = new HashMap<String,Integer>();
        for(int current = 0; current < text.length; current++) {
            String token = text[current].toLowerCase();
            if(token.length() >= mintokenlen) {
                // First time we see a specific token
                if(_lastPosition.get(token) == null) {
                	_totalDistance.put(token, new Integer(0));
                	_lastPosition.put(token, new Integer(current));
                } else {
                	// n:th (n>1) time we see a specific token
                	int distance = current - _lastPosition.get(token).intValue();
                	int totalDistance = distance + _totalDistance.get(token).intValue();
                	_totalDistance.put(token, new Integer(totalDistance));
                	_lastPosition.put(token, new Integer(current));
                	// First distance for a specific token
                	if(_distances.get(token) == null) {
                    	ArrayList<Integer> distances = new ArrayList<Integer>();
                    	distances.add(new Integer(distance));
                    	_distances.put(token, distances);
                    } else {
                    	// n:th (n>1) plus distance for a specific token
                    	_distances.get(token).add(new Integer(distance));
                    }
                }
				counted++;
            } else {
            	_standardDeviations.put(token, new Double(1.0));
            }
        }
        // Calculate the standard deviation of the above stored distances, see 
        // "Keyword detection in natural languages and DNA" (Ortuno et al 2002)
        for(Iterator iter = _distances.keySet().iterator(); iter.hasNext();) {
			String token = (String)iter.next();
			double stotal = 0, s2total = 0;
			double meanx = (_totalDistance.get(token) / _distances.get(token).size());
			for(int i = 0; i < _distances.get(token).size(); i++) {
				int x = _distances.get(token).get(i);
				double s = x / meanx;
				s2total += Math.pow(s, 2);
				stotal += s;
			}
			double s2mean = s2total / _distances.get(token).size();
			double smean2 = Math.pow(stotal / _distances.get(token).size(), 2);
			double sigma = Math.sqrt(s2mean / smean2);
			_standardDeviations.put(token, new Double(sigma));
		}
		return counted;
	}

	/**
	 * Usage: moj.lang.StandardDeviationList &lt;file&gt; (&lt;minimum token length&gt;)<br />
	 * &lt;file&gt; : file to build standard deviations on<br />
	 * &lt;minimum token length&gt; : minimum length for a token for it to be counted<br />
	 */
	public static void main(String[] args) {
        if(args.length == 0) {
    		System.out.println("*** StandardDeviationList ***");
    		System.out.println("Usage: moj.lang.StandardDeviationList <file> (<minimum token length>)");
    		System.out.println("<file> : file to build standard deviations on");
    		System.out.println("<minimum token length> : minimum length for a token for it to be counted");
    		System.exit(0);
    	}
        int mintokenlen = 0;
        if(args.length > 1)
        	mintokenlen = Integer.parseInt(args[1]);
        // Build list of standard deviation of distances between unique tokens
		GenericTokenizer gt = new GenericTokenizer();
		StandardDeviationList sdl = new StandardDeviationList();
		sdl.addText(gt.getWords(FileTools.readFile2String(args[0])), mintokenlen);
		// Print sorted deviation list
		List<Map.Entry> entries = new ArrayList<Map.Entry>(sdl.getStandardDeviations().entrySet());
		Collections.sort(entries, new EntryValueComparator());
        for(int i = 0; i < entries.size(); i++) {
            Map.Entry entry = entries.get(i);
        	if(((String)entry.getKey()).length() > 0)
            	System.out.println(entry.getKey() + "\t" + entry.getValue());
        }
	}
}
