/*
 * MartinsWS.java
 * Copyright (c) 2004, KTH NADA.
 *
 * This file is part of SweSum^2 (see http://swesum.nada.kth.se/),
 * and is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; see the file COPYING.  If not, see
 * http://opensource.org/licenses/gpl-license.php
 *
 * Martin Hassel, 2004-May-08
 * http://www.nada.kth.se/~xmartin/
 */

package moj.ri.weighting;

import moj.ri.RandomLabel;

/**
 * Calculates the weighting based upon the distance to the current label in 
 * the following manner: weight=(1/distance to focus word)
 * but also gives a higher weight to content words. The index term 
 * (<code>word</code>) should be in the form <i>word_pos</i> for this 
 * WeightingScheme to work properly. 
 * 
 * @author  Martin Hassel
 * @version 2004-Nov-23
 */
public class MartinsWS extends WeightingScheme {
	private static final long serialVersionUID = -8256022455724831455L;

	/**
	 * Calculates the weight based upon the distance (in the left context) to
	 * the current label in the following manner: weight=(1/distance to focus word)
	 * @param focusLabel RandomLabel in focus, i.e. the focus word whose context vector 
	 *        should be modified with the weighted RandomLabel in left context.
	 * @param distance leftward distance to the focus word (RandomLabel).
	 * @param leftContextLabel RandomLabel in left context at <code>distance</code> from 
	 *        the focus word, which can be used together with <code>distance</code> to 
	 *        calculate the weight.
	 * @return the context of the focus word modified according to the weighting scheme.
	 */
	public float[] applyLeftWeighting(RandomLabel focusLabel, int distance, RandomLabel leftContextLabel) {
		float[] focusContext = focusLabel.getContext();
		byte weight = 1;
		if(focusLabel.getWord().matches(".*_[jj|nn|pn|pm|vb]$") && leftContextLabel.getWord().matches(".*_[jj|nn|pn|pm|vb]$"))
			weight = 2;
		// First we update the negative positions (-1:s)
		int[] negs = leftContextLabel.getNegativePositions();
		for(int current = 0; current < negs.length; current++)
		    focusContext[negs[current]] += ((-1/distance)*weight);
		// Then we update the positive positions (1:s)
		int[] poss = leftContextLabel.getPositivePositions();
		for(int current = 0; current < poss.length; current++)
		    focusContext[poss[current]] += ((1/distance)*weight);
		return focusContext;
	}

	/**
	 * Calculates the weight based upon the distance (in the right context) to
	 * the current label in the following manner: weight=(1/distance to focus word)
	 * @param focusLabel RandomLabel in focus, i.e. the focus word whose context vector 
	 *        should be modified with the weighted RandomLabel in right context.
	 * @param distance rightward distance to the focus word (RandomLabel).
	 * @param rightContextLabel RandomLabel in right context at <code>distance</code> from 
	 *        the focus word, which can be used together with <code>distance</code> to 
	 *        calculate the weight.
	 * @return the context of the focus word modified according to the weighting scheme.
	 */
	public float[] applyRightWeighting(RandomLabel focusLabel, int distance, RandomLabel rightContextLabel) {
		float[] focusContext = focusLabel.getContext();
		byte weight = 1;
		if(focusLabel.getWord().matches(".*_[jj|nn|pn|pm|vb]$") && rightContextLabel.getWord().matches(".*_[jj|nn|pn|pm|vb]$"))
			weight = 2;
		// First we update the negative positions (-1:s)
		int[] negs = rightContextLabel.getNegativePositions();
	    for(int current = 0; current < negs.length; current++)
	    	focusContext[negs[current]] += ((-1/distance)*weight);
		// Then we update the positive positions (1:s)
	    int[] poss = rightContextLabel.getPositivePositions();
	    for(int current = 0; current < poss.length; current++)
	    	focusContext[poss[current]] += ((1/distance)*weight);
		return focusContext;
	}
}
