/*
 * WeightingScheme.java
 * Copyright (c) 2004, KTH NADA.
 *
 * This file is part of SweSum^2 (see http://swesum.nada.kth.se/),
 * and is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; see the file COPYING.  If not, see
 * http://opensource.org/licenses/gpl-license.php
 *
 * Martin Hassel, 2004-May-08
 * http://www.nada.kth.se/~xmartin/
 */

package moj.ri.weighting;

import java.io.Serializable;

import moj.ri.RandomLabel;

/**
 * Calculates the weighting of the context label. <br />
 * All subclasses of <code>WeightingScheme</code> <i>must</i> provide methods 
 * for calculating the weights for left and right context windows according 
 * to the desired algorithm (for example using distance to context label).
 * If there is a need to be able to serialize the <code>RandomIndex</code>, all 
 * subclasses of <code>WeightingScheme</code> <i>must</i> declare their own 
 * <code>serialVersionUID</code>.
 * 
 * @author  Martin Hassel
 * @version 2005-Apr-04
 */
public abstract class WeightingScheme implements Serializable {
	private static final long serialVersionUID = -6781079375065063171L;

	/**
	 * Adds the weighted context label (in the left context) to the focus label's 
	 * context label and returns it. 
	 * Application writers <i>must</i> its own methods in a subclass to calculate 
	 * the weights according to specific algorithms.
	 * @param focusLabel RandomLabel in focus, i.e. the focus word whose context vector 
	 *        should be modified with the weighted RandomLabel in left context.
	 * @param distance leftward distance to the focus word (RandomLabel).
	 * @param leftContextLabel RandomLabel in left context at <code>distance</code> from 
	 *        the focus word, which can be used together with <code>distance</code> to 
	 *        calculate the weight.
	 * @return the context of the focus word modified according to the weighting scheme.
	 */
	public abstract float[] applyLeftWeighting(RandomLabel focusLabel, int distance, RandomLabel leftContextLabel);

	/**
	 * Adds the weighted context label (in the right context) to the focus label's 
	 * context label and returns it. 
	 * Application writers <i>must</i> its own methods in a subclass to calculate 
	 * the weights according to specific algorithms.
	 * @param focusLabel RandomLabel in focus, i.e. the focus word whose context vector 
	 *        should be modified with the weighted RandomLabel in right context.
	 * @param distance rightward distance to the focus word (RandomLabel).
	 * @param rightContextLabel RandomLabel in right context at <code>distance</code> from 
	 *        the focus word, which can be used together with <code>distance</code> to 
	 *        calculate the weight.
	 * @return the context of the focus word modified according to the weighting scheme.
	 */
	public abstract float[] applyRightWeighting(RandomLabel focusLabel, int distance, RandomLabel rightContextLabel);
}
